---
--[[ 
This page will provide documentation for the syntax of a "rotation definition 
file". A rotation definition file is valid Lua, and therefore all of the
control structures and other language features are valid in a definition file.
However, there are some extra functions that have been added in order to make
life a little bit easier for people with limited knowledge of Lua. This page 
will show you the syntax of these extra functions and what they mean.

All definition files follow a basic layout and this layout must be followed.
@class file
--]]

---
--[[
Checks whether or not the player is in combat based on whether or not the 
UI is locked down.
@return 1 if the player is in combat, nil otherwise (1nil)
--]]
local function InCombat()
    return InCombatLockdown(); 
end;

---
--[[ 
Checks to see whether or not the player has a pet that shows the pet bar,
such as Unholy Death Knight ghouls or warlock demons.
@return 1 if the player has a pet, nil otherwise (1nil)
--]] 
local function HasPet()
    return UnitExists();
end;

---
--[[
Checks the health of the player's pet.
@return Percentage health of the player's pet, 0 if it is dead, and nil if it doesn't exist.
--]]
local function PetHealth()
    if (UnitExists("pet") and not UnitIsDead("pet")) then
        local currHP = UnitHealth("pet");
        local maxHP = UnitHealthMax("pet");
        local result = math.ceil((currHP / maxHP) * 100);
    elseif (UnitIsDead("pet")) then
        local result = 0;
    else
        local result = nil;
    end;
    return result;
end;

---
--[[
Checks whether or not the player's target has the specified buff.
@param name The name of the buff being checked for.
@param checkme OPTIONAL If we should check whether or not the player casted 
the spell.
@return 1 if the target has the buff (and it was casted by the player if
checkme was not nil), nil otherwise. (1nil).
--]]
local function TargetHasBuff(name, checkme)
    if (UnitExists("target")) then
        local bname, _, _, _, _, _, _, _, _ = UnitBuff("target", name);
    end;
    if (bname) then
        return 1;
    else 
        return nil;
    end;
end;

---
--[[
Checks whether or not the player's target has the specified debuff.
@param name The name of the buff being checked for.
@param checkme OPTIONAL If we should check whether or not the player casted 
the spell.
@return 1 if the target has the debuff (and it was casted by the player if
checkme was not nil), nil otherwise. (1nil)
--]]
local function TargetHasDebuff(name, checkme)
    if (UnitExists("target")) then
        local bname, _, _, _, _, _, _, caster, _ = UnitDebuff("target", name);
    end;
    if checkme then
        if (bname == name and caster == player) then
            return 1;
        else
            return nil;
        end;
    else
        if (bname == name) then
            return 1;
        else
            return nil;
        end;
    end;
end;

---
--[[
Checks whether or not the player's target has the specified debuff AND it was
cast by the player. This is an alias for TargetHasDebuff(name, 1).
@param name The debuff to be checked.
@return 1 if the target has the debuff and it was cast by the player, nil
otherwise. (1nil)
--]]
local function TargetHasDebuffByMe(name)
    return TargetHasDebuff(name, 1);
end;

---
--[[
Checks whether or not the player has the specified buff.
@param name The name of the buff being checked for.
@return 1 if the target has the debuff, nil otherwise. (1nil)
--]]
local function PlayerHasBuff()
end;

local function PlayerHasBuffByMe()
    return PlayerHasBuff(name, 1);
end;

---
--[[
Checks whether or not the player has the specified debuff.
@param name The name of the debuff being checked for.
@return 1 if the target has the debuff, nil otherwise (1nil)
--]]
local function PlayerHasDebuff()
end;

---
--[[
Checks if the player can cast the specified spell.
@param name The name of the spell being checked.
@return 1 if the spell is castable, nil otherwise (1nil)
--]]
local function Castable(name)
    local castable, mana = IsUsableSpell(spell);
    if (not castable or mana == 1) then
        return nil;
    end
    local start, duration, enabled = GetSpellCooldown(spell);
    if (duration == 0) then
        return 1;
    else
        return nil;
    end
end;

---
--[[
Checks whether or not the player has the specified glyph in their current
talent spec.
@param name The name of the glyph being checked for without the "Glyph of"
in front of it, (i.e. Glyph of Souls -> "Souls")
@return 1 if the glyph is inscribed, nil otherwise. (1nil)
--]]
local function HasGlyph()

end;

local function Alternative()
end;

local function Cliptime()
end;

local function KnowSpell()
end;

local function HasTalent()
end;

local function OnGCD()
end;

local function LastCast()
end;

local function PetExists()
end;

local function TargetHealth()
end;

local function PlayerHealth()
end;

local function Threat()
end;

local function TargetType()
end;

local function IsBoss()
end;

local function MxBuff()
end;

local function MxDebuff()
end;

local function MxBuffByMe()
end;

local function MxDebuffByMe()
end;